def prevent_long_paths(path: str, max_length: int = 512) -> str:
    """
    Prevents the path from being too long.

    Default max length is 512 characters.
    """
    max_ = max(2, max_length if max_length > 0 else 512)

    if len(path) <= max_:
        return path

    half = max_ // 2
    return f"{path[:half]}...{path[-half:]}"



"""
This module contains the different store classes that can be used to store the traffic data.
"""

from .csv_store import CSVStore
from .json_store import JSONStore
from .sql_orm_store import SQLORMStore, SQLORMModelMixin
from .sql_store import SQLStore
from .redis_store import RedisStore

__all__ = [
    "JSONStore",
    "CSVStore",
    "SQLStore",
    "SQLORMStore",
    "SQLORMModelMixin",
    "RedisStore",
]



from typing import Optional, Iterable


class LogPolicy:
    """
    The LogPolicy class is used to define what data should be logged
    when a request is made to the Flask app.

    If a LogPolicy is not passed to a store, one is created and all attributes
    are set to True.
    """

    request_date: bool
    request_method: bool
    request_host_url: bool
    request_path: bool
    request_endpoint: bool
    request_remote_address: bool
    request_referrer: bool
    request_user_agent: bool
    request_browser: bool
    request_platform: bool

    response_time: bool
    response_size: bool
    response_status_code: bool
    response_exception: bool
    response_mimetype: bool

    on_endpoints: Iterable
    skip_endpoints: Iterable

    on_status_codes: Iterable
    skip_status_codes: Iterable

    only_on_exception: bool
    skip_on_exception: bool

    max_request_path_length: int

    def __repr__(self) -> str:
        return f"<LogPolicy {self.__dict__}>"

    def __init__(
        self,
        on_endpoints: Optional[Iterable[str]] = None,
        skip_endpoints: Optional[Iterable[str]] = None,
        on_status_codes: Optional[Iterable[str]] = None,
        skip_status_codes: Optional[Iterable[str]] = None,
        only_on_exception: bool = False,
        skip_on_exception: bool = False,
        max_request_path_length: int = 512,
    ) -> None:
        """
        Create a new LogPolicy instance.

        Use either `.set_from_true` or `.set_from_false` methods
        to enable or disable logging of data.

        ::

            log_policy = LogPolicy()
            log_policy.set_from_true(...)
            -or-
            log_policy.set_from_false(...)

        *Default Policy:*

        ::

            request_method = True
            request_host_url = True
            request_path = True
            request_endpoint = True
            request_remote_address = True
            request_referrer = True
            request_user_agent = True
            request_browser = True
            request_platform = True

            response_time = True
            response_size = True
            response_status_code = True
            response_exception = True
            response_mimetype = True

            on_endpoints = None (disabled)
            skip_endpoints = None (disabled)

            on_status_codes = None (disabled)
            skip_status_codes = None (disabled)

            only_on_exception = False
            skip_on_exception = False

            max_request_path_length: int = 512

        :param on_endpoints: only log requests to these endpoints
        :param skip_endpoints: do not log requests to these endpoints
        :param on_status_codes: only log requests with these status codes
        :param skip_status_codes: do not log requests with these status codes
        :param only_on_exception: only create a log entry if an exception is raised during the request if True
        :param skip_on_exception: do not create a log entry if an exception is raised during the request if True
        :param max_request_path_length: the maximum length of the request path to log
        """

        self.request_date = True
        self.request_method = True
        self.request_host_url = True
        self.request_path = True
        self.request_endpoint = True
        self.request_remote_address = True
        self.request_referrer = True
        self.request_user_agent = True
        self.request_browser = True
        self.request_platform = True

        self.response_time = True
        self.response_size = True
        self.response_status_code = True
        self.response_exception = True
        self.response_mimetype = True

        # Endpoints
        if on_endpoints is None:
            self.on_endpoints = set()
        else:
            self.on_endpoints = on_endpoints

        if skip_endpoints is None:
            self.skip_endpoints = set()
        else:
            self.skip_endpoints = skip_endpoints

        # Status Codes
        if on_status_codes is None:
            self.on_status_codes = set()
        else:
            self.on_status_codes = on_status_codes

        if skip_status_codes is None:
            self.skip_status_codes = set()
        else:
            self.skip_status_codes = skip_status_codes

        self.only_on_exception = only_on_exception
        self.skip_on_exception = skip_on_exception

        self.max_request_path_length = max_request_path_length

    def set_from_true(
        self,
        request_date: bool = True,
        request_method: bool = True,
        request_host_url: bool = True,
        request_path: bool = True,
        request_endpoint: bool = True,
        request_remote_address: bool = True,
        request_referrer: bool = True,
        request_user_agent: bool = True,
        request_browser: bool = True,
        request_platform: bool = True,
        response_time: bool = True,
        response_size: bool = True,
        response_status_code: bool = True,
        response_exception: bool = True,
        response_mimetype: bool = True,
    ) -> "LogPolicy":
        """
        Disable what you don't want to log.

        Set the attribute you don't want to log to False.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        self.request_date = request_date
        self.request_method = request_method
        self.request_host_url = request_host_url
        self.request_path = request_path
        self.request_endpoint = request_endpoint
        self.request_remote_address = request_remote_address
        self.request_referrer = request_referrer
        self.request_user_agent = request_user_agent
        self.request_browser = request_browser
        self.request_platform = request_platform

        self.response_time = response_time
        self.response_size = response_size
        self.response_status_code = response_status_code
        self.response_exception = response_exception
        self.response_mimetype = response_mimetype

        return self

    def set_from_false(
        self,
        request_date: bool = False,
        request_method: bool = False,
        request_host_url: bool = False,
        request_path: bool = False,
        request_endpoint: bool = False,
        request_remote_address: bool = False,
        request_referrer: bool = False,
        request_user_agent: bool = False,
        request_browser: bool = False,
        request_platform: bool = False,
        response_time: bool = False,
        response_size: bool = False,
        response_status_code: bool = False,
        response_exception: bool = False,
        response_mimetype: bool = False,
    ) -> "LogPolicy":
        """
        Enable what you want to log.

        Set the attribute you want to log to True.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        self.request_date = request_date
        self.request_method = request_method
        self.request_host_url = request_host_url
        self.request_path = request_path
        self.request_endpoint = request_endpoint
        self.request_remote_address = request_remote_address
        self.request_referrer = request_referrer
        self.request_user_agent = request_user_agent
        self.request_browser = request_browser
        self.request_platform = request_platform

        self.response_time = response_time
        self.response_size = response_size
        self.response_status_code = response_status_code
        self.response_exception = response_exception
        self.response_mimetype = response_mimetype

        return self



import csv
import typing as t
from datetime import datetime
from pathlib import Path

from .._globals import IGNORE_LOCALS
from .._helpers import prevent_long_paths
from .._log_policy import LogPolicy

if t.TYPE_CHECKING:
    from .._traffic import Traffic


class CSVStore:
    filename: str
    location: t.Optional[t.Union[str, Path]]

    filepath: Path
    log_policy: LogPolicy

    _traffic_instance = None

    def __repr__(self) -> str:
        return f"<CSVStore filename={self.filepath} log_policy={self.log_policy}>"

    def __init__(
        self,
        filename: str = "traffic.csv",
        location: t.Optional[t.Union[str, Path]] = None,
        log_policy: LogPolicy = None,
    ) -> None:
        """
        Create a new CSVStore instance.

        :param filename: the name of the CSV file to store the traffic data in
        :param location: the location of the CSV file
        :param log_policy: the log policy to use (defaults to log everything if not provided)
        """
        self.filename = filename
        self.location = location

        if log_policy is None:
            from .._log_policy import LogPolicy

            self.log_policy = LogPolicy()

        else:
            self.log_policy = log_policy

    def setup(self, traffic_instance: "Traffic") -> None:
        """
        Set up the CSVStore instance.

        :param traffic_instance:
        :return:
        """
        # set filepath to instance folder if location is None
        if self.location is None:
            self.filepath = traffic_instance.app_instance_folder / self.filename
        else:
            # This expects an absolute path
            if isinstance(self.location, str):
                self.filepath = Path(self.location) / self.filename

            # This expects a Path object
            if isinstance(self.location, Path):
                self.filepath = self.location / self.filename

        # create the file if it doesn't exist
        if not self.filepath.exists():
            # create the parent directory if it doesn't exist
            if not self.filepath.parent.exists():
                self.filepath.parent.mkdir(parents=True)

            # file is created here
            self.filepath.touch()

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None:
        """
        Log the traffic data.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        data = {}

        for attr, attr_val in self.log_policy.__dict__.items():
            if attr in IGNORE_LOCALS:
                continue

            if attr_val:
                # Prevent long paths from being stored
                if attr == "request_path":
                    data[attr] = prevent_long_paths(
                        request_path, self.log_policy.max_request_path_length
                    )
                    continue

                if isinstance(locals()[attr], datetime):
                    data[attr] = locals()[attr].isoformat()
                    continue

                data[attr] = locals()[attr]

            else:
                data[attr] = None

        with open(self.filepath, "a+") as f:
            writer = csv.DictWriter(f, fieldnames=data.keys())  # type: ignore

            if f.tell() == 0:
                # write the header if the file is empty
                writer.writeheader()

            writer.writerow(data)

    def read(self) -> t.List[t.Dict[str, t.Any]]:
        """
        Read the CSV file and return the data as a list of dictionaries.

        :return: the data as a list of dictionaries
        """
        with open(self.filepath, "r") as f:
            reader = csv.DictReader(f)

            return list(reversed(list(reader)))



import typing as t
from datetime import datetime

if t.TYPE_CHECKING:
    from flask_traffic import Traffic
    from flask_traffic._log_policy import LogPolicy


class StoreProtocol(t.Protocol):
    """
    The protocol to follow for a valid *Store.
    """

    log_policy: "LogPolicy"

    def setup(self, traffic_instance: "Traffic") -> None: ...

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None: ...

    def read(self) -> t.Any: ...



import typing as t
from datetime import datetime
from pathlib import Path

from .._helpers import prevent_long_paths

try:
    from sqlalchemy import (
        create_engine,
        Table,
        Column,
        String,
        Integer,
        DateTime,
        MetaData,
        BigInteger,
    )
except ImportError:
    raise ImportError(
        "You're attempting to use a SQLStore but sqlalchemy is not installed. "
        "Please install it with `pip install sqlalchemy`"
    )

from .._globals import IGNORE_LOCALS
from .._log_policy import LogPolicy

if t.TYPE_CHECKING:
    from .._traffic import Traffic
    from sqlalchemy.engine import Engine


class SQLStore:
    filename: str
    location: t.Optional[t.Union[str, Path]]

    filepath: t.Optional[Path]
    log_policy: LogPolicy

    database_table_name: t.Optional[str]
    database_url: t.Optional[str]
    database_engine: t.Optional["Engine"]
    database_metadata: t.Optional["MetaData"]
    database_log_table: t.Optional["Table"]

    _traffic_instance = None

    def __repr__(self) -> str:
        if self.database_url:
            return f"<SQLStore database_url={self.database_url} log_policy={self.log_policy}>"

        if self.database_engine:
            return f"<SQLStore database_engine={self.database_engine} log_policy={self.log_policy}>"

        return f"<SQLStore filename={self.filepath} log_policy={self.log_policy}>"

    def __init__(
        self,
        filename: t.Optional[str] = "traffic.sqlite",
        location: t.Optional[t.Union[str, Path]] = None,
        log_policy: LogPolicy = None,
        *,
        database_table_name: t.Optional[str] = "_traffic_",
        database_url: t.Optional[str] = None,
        database_engine: t.Optional["Engine"] = None,
    ) -> None:
        """
        Create a new SQLStore instance.

        :param filename: the name of a SQLite database file (skipped if database_url or database_engine is provided)
        :param location: the location of the SQLite database file  (skipped if database_url or database_engine is provided)
        :param log_policy: the log policy to use (defaults to log everything if not provided)
        :param database_table_name: the name of the table to store the traffic data in
        :param database_url: the URL of the database to connect to
        :param database_engine: the SQLAlchemy engine to use (skipped if database_url is provided)
        """
        if log_policy is None:
            from .._log_policy import LogPolicy

            self.log_policy = LogPolicy()

        else:
            self.log_policy = log_policy

        self.database_table_name = database_table_name

        if database_url is not None:
            self.database_url = database_url
            return
        else:
            self.database_url = None

        if database_engine is not None:
            self.database_engine = database_engine
            return
        else:
            self.database_engine = None

        self.filename = filename
        self.location = location

    def _build_database_filepath(self, traffic_instance: "Traffic") -> None:
        """
        If database is determined to be a SQLite database, build the filepath.

        :param traffic_instance:
        :return:
        """
        if self.location is None:
            self.filepath = traffic_instance.app_instance_folder / self.filename
        else:
            if isinstance(self.location, str):
                self.filepath = Path(self.location) / self.filename

            if isinstance(self.location, Path):
                self.filepath = self.location / self.filename

    def _init_database(
        self, url: t.Optional[str] = None, engine: t.Optional["Engine"] = None
    ) -> None:
        """
        Initialize the database connection and create the table.

        :param url:
        :param engine:
        :return:
        """
        if url:
            self.database_engine = create_engine(self.database_url)
        elif engine:
            self.database_engine = engine
        else:
            self.database_engine = create_engine(f"sqlite:///{self.filepath}")

        self.database_metadata = MetaData()
        self.database_log_table = Table(
            self.database_table_name,
            self.database_metadata,
            Column("traffic_id", BigInteger, primary_key=True),
            Column("request_date", DateTime, nullable=True),
            Column("request_method", String, nullable=True),
            Column("request_host_url", String, nullable=True),
            Column("request_path", String, nullable=True),
            Column("request_endpoint", String, nullable=True),
            Column("request_remote_address", String, nullable=True),
            Column("request_referrer", String, nullable=True),
            Column("request_user_agent", String, nullable=True),
            Column("request_browser", String, nullable=True),
            Column("request_platform", String, nullable=True),
            Column("response_time", Integer, nullable=True),
            Column("response_size", String, nullable=True),
            Column("response_status_code", Integer, nullable=True),
            Column("response_exception", String, nullable=True),
            Column("response_mimetype", String, nullable=True),
        )

    def setup(self, traffic_instance: "Traffic") -> None:
        """
        Set up the SQLStore instance.

        :param traffic_instance:
        :return:
        """
        if self.database_url:
            self._init_database(url=self.database_url)

        elif self.database_engine:
            self._init_database(engine=self.database_engine)

        else:
            self._build_database_filepath(traffic_instance)
            self._init_database()

        self.database_metadata.create_all(self.database_engine)

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None:
        """
        Log the traffic data.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        data = {}

        for attr, attr_val in self.log_policy.__dict__.items():
            if attr in IGNORE_LOCALS:
                continue

            if attr_val:
                # Prevent long paths from being stored
                if attr == "request_path":
                    data[attr] = prevent_long_paths(
                        request_path, self.log_policy.max_request_path_length
                    )
                    continue

                data[attr] = locals()[attr]

            else:
                data[attr] = None

        with self.database_engine.connect() as connection:
            connection.execute(self.database_log_table.insert().values(data))
            connection.commit()

    def read(self, limit: int = 10000) -> t.List[t.Dict[str, t.Any]] | None:
        with self.database_engine.connect() as connection:
            sel = self.database_log_table.select().order_by(
                self.database_log_table.c.traffic_id.desc()
            )
            if limit:
                sel = sel.limit(limit)

            results = connection.execute(sel)

            return [row._asdict() for row in results.fetchall()]



import typing as t
from datetime import datetime

from .._helpers import prevent_long_paths



from .._globals import IGNORE_LOCALS
from .._log_policy import LogPolicy

if t.TYPE_CHECKING:
    from .._traffic import Traffic
    from sqlalchemy.orm import Session
    from sqlalchemy.sql._typing import _DMLTableArgument

    try:
        from sqlalchemy import Column, String, Integer, DateTime, BigInteger
        from sqlalchemy import insert, select, desc
    except ImportError:
        raise ImportError(
            "You're attempting to use a SQLORMStore but sqlalchemy is not installed. "
            "Please install it with `pip install sqlalchemy` or `pip install flask-sqlalchemy`"
        )


class SQLORMModelMixin:
    """
    A mixin for use with ORM SQLAlchemy / Flask-SQLAlchemy models.

    Change the table name by setting the __tablename__ attribute
    in your model, if needed.

    Do not change the column names or types, as these are used
    by flask-traffic internally.

    *Flask-SQLAlchemy example:*

    ::

        class MyTrafficModel(db.Model, SQLORMModelMixin):
            __tablename__ = "my_traffic_table"

    *Defaults:*

    ::

        __tablename__ = "_traffic_"
        traffic_id = Column(BigInteger, primary_key=True)
        request_date = Column(DateTime, nullable=True)
        request_method = Column(String, nullable=True)
        request_host_url = Column(String, nullable=True)
        request_path = Column(String, nullable=True)
        request_endpoint = Column(String, nullable=True)

        request_remote_address = Column(String, nullable=True)
        request_referrer = Column(String, nullable=True)
        request_user_agent = Column(String, nullable=True)
        request_browser = Column(String, nullable=True)
        request_platform = Column(String, nullable=True)
        response_time = Column(Integer, nullable=True)
        response_size = Column(String, nullable=True)
        response_status_code = Column(Integer, nullable=True)
        response_exception = Column(String, nullable=True)
        response_mimetype = Column(String, nullable=True)

    """

    __tablename__ = "_traffic_"
    traffic_id = Column(BigInteger, primary_key=True)
    request_date = Column(DateTime, nullable=True)
    request_method = Column(String, nullable=True)
    request_host_url = Column(String, nullable=True)
    request_path = Column(String, nullable=True)
    request_endpoint = Column(String, nullable=True)
    request_remote_address = Column(String, nullable=True)
    request_referrer = Column(String, nullable=True)
    request_user_agent = Column(String, nullable=True)
    request_browser = Column(String, nullable=True)
    request_platform = Column(String, nullable=True)
    response_time = Column(Integer, nullable=True)
    response_size = Column(String, nullable=True)
    response_status_code = Column(Integer, nullable=True)
    response_exception = Column(String, nullable=True)
    response_mimetype = Column(String, nullable=True)


class SQLORMStore:
    db_session: t.Optional["Session"]
    model: "_DMLTableArgument"

    def __repr__(self) -> str:
        return f"<SQLORMStore model={self.model.__name__} log_policy={self.log_policy}>"

    def __init__(
        self,
        model: "_DMLTableArgument",
        db_session: t.Optional["Session"] = None,
        log_policy: LogPolicy = None,
    ) -> None:
        """
        Create a new SQLORMStore instance.

        :param model: the SQLAlchemy model to insert logs into
        :param db_session: the SQLAlchemy session to use (this is automatically set if using Flask-SQLAlchemy)
        :param log_policy: the log policy to use (defaults to log everything if not provided)
        """
        if log_policy is None:
            from .._log_policy import LogPolicy

            self.log_policy = LogPolicy()

        self.model = model
        self.db_session = db_session

    def setup(self, traffic_instance: "Traffic") -> None:
        """
        Set up the SQLORMStore instance.

        :param traffic_instance:
        :return:
        """
        if self.model is None:
            raise ValueError("No model was passed in. ORMStore(..., model: Model)")

        if self.db_session is None:
            if traffic_instance.app.extensions.get("sqlalchemy") is None:
                raise ImportError(
                    "No SQLAlchemy session was found or passed in. "
                    "ORMStore(..., db_session=session)"
                )
            try:
                self.db_session = traffic_instance.app.extensions["sqlalchemy"].session
            except KeyError:
                raise ImportError(
                    "No SQLAlchemy session was found or passed in. "
                    "ORMStore(..., db_session=session)"
                )

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None:
        """
        Log the traffic data.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        data = {}

        for attr, attr_val in self.log_policy.__dict__.items():
            if attr in IGNORE_LOCALS:
                continue

            if attr_val:
                # Prevent long paths from being stored
                if attr == "request_path":
                    data[attr] = prevent_long_paths(
                        request_path, self.log_policy.max_request_path_length
                    )
                    continue

                data[attr] = locals()[attr]

            else:
                data[attr] = None

        self.db_session.execute(insert(self.model).values(data))
        self.db_session.commit()

    def read(self, limit: int = 10000) -> list[dict[str, t.Any]] | None:
        if hasattr(self.model, "traffic_id"):
            sel = select(self.model).order_by(desc(self.model.traffic_id))
            if limit:
                sel = sel.limit(limit)

            results = self.db_session.execute(sel)

            if results:
                return [
                    {
                        "request_date": getattr(row, "request_date", None),
                        "request_method": getattr(row, "request_method", None),
                        "request_host_url": getattr(row, "request_host_url", None),
                        "request_path": getattr(row, "request_path", None),
                        "request_endpoint": getattr(row, "request_endpoint", None),
                        "request_remote_address": getattr(
                            row, "request_remote_address", None
                        ),
                        "request_referrer": getattr(row, "request_referrer", None),
                        "request_user_agent": getattr(row, "request_user_agent", None),
                        "request_browser": getattr(row, "request_browser", None),
                        "request_platform": getattr(row, "request_platform", None),
                        "response_time": getattr(row, "response_time", None),
                        "response_size": getattr(row, "response_size", None),
                        "response_status_code": getattr(
                            row, "response_status_code", None
                        ),
                        "response_exception": getattr(row, "response_exception", None),
                        "response_mimetype": getattr(row, "response_mimetype", None),
                    }
                    for row in results.scalars().all()
                ]

            return []

        return None



import typing as t
from datetime import datetime

from .._helpers import prevent_long_paths


from .._globals import IGNORE_LOCALS
from .._log_policy import LogPolicy

if t.TYPE_CHECKING:
    from .._traffic import Traffic
    try:
        from redis import Redis
    except ImportError:
        raise ImportError(
            "You're attempting to use a RedisStore but redis is not installed. "
            "Please install it with `pip install redis`"
        )


class RedisStore:
    client: "Redis"
    log_policy: LogPolicy
    event_name: str

    _traffic_instance = None

    def __repr__(self) -> str:
        return f"<RedisStore client={self.client} log_policy={self.log_policy}>"

    def __init__(
        self,
        redis_host: str,
        event_name: str,
        redis_password: str | None = None,
        redis_port: int = 6379,
        log_policy: LogPolicy | None = None,
    ) -> None:
        try:
            from redis import Redis
        except ImportError:
            raise ImportError(
                "You're attempting to use a RedisStore but redis is not installed. "
                "Please install it with `pip install redis`"
            )
        """
        Create a new RedisStore instance.

        :param redis_host: redis host to connect to
        :param event_name: redis stream name to use for logging traffic data.
        :param redis_password: redis password for authentication
        :param redis_port: redis port on which redis instance is listening on
        :param log_policy: the log policy to use (defaults to log everything if not provided)
        """
        self.client = Redis(
            host=redis_host,
            port=redis_port,
            password=redis_password,
            db=0,
            decode_responses=True,
        )
        self.event_name = event_name
        if log_policy is None:
            from .._log_policy import LogPolicy

            self.log_policy = LogPolicy()
        else:
            self.log_policy = log_policy

    def setup(self, traffic_instance: "Traffic"):
        """
        Nothing to set up in redis store
        """
        pass

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None:
        """
        Log the traffic data.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        legal_types = (str, int, float, bytes)

        data = {}

        for attr, attr_val in self.log_policy.__dict__.items():
            if attr in IGNORE_LOCALS:
                continue

            if attr_val:
                # Prevent long paths from being stored
                if attr == "request_path":
                    data[attr] = prevent_long_paths(
                        request_path, self.log_policy.max_request_path_length
                    )
                    continue

                if isinstance(locals()[attr], datetime):
                    data[attr] = locals()[attr].isoformat()
                    continue

                if locals()[attr] is None:
                    data[attr] = "None"
                    continue

                if not isinstance(locals()[attr], legal_types):
                    data[attr] = "unknown"
                else:
                    data[attr] = locals()[attr]

            else:
                continue

        self.client.xadd(name=self.event_name, fields=data)

    def read(self) -> t.List[t.Dict[str, t.Any]]:
        """
        Read the Redis stream and return the contents.

        :return: the contents of the redis stream
        """

        logs: t.List[t.Dict[str, t.Any]] = []

        data_stream = self.client.xread(streams={self.event_name: 0})
        if len(data_stream) > 0:
            msg_stream = data_stream[0]
            messages = msg_stream[1]

            for msg in messages:
                logs.append(
                    {
                        key: (None if value == "None" else value)
                        for key, value in msg[1].items()
                    }
                )

        logs.reverse()
        return logs
